#include "./alpm/manager.h"
#include "./alpm/utilities.h"
#include "./alpm/config.h"
#include "./alpm/resolvebuildorder.h"
#include "./alpm/mingwbundle.h"

#include "./network/server.h"

#include <c++utilities/application/argumentparser.h>
#include <c++utilities/conversion/stringconversion.h>
#include <c++utilities/application/failure.h>

#include <QCoreApplication>

#include <iostream>
#include <algorithm>

using namespace std;
using namespace ApplicationUtilities;
using namespace RepoIndex;

int main(int argc, char *argv[])
{
    // check whether shell syntax is enabled
    for(char **arg = argv, **end = argv + argc; arg != end; ++arg) {
        if(strcmp(*arg, "--sh-syntax") == 0) {
            useShSyntax = true;
            shchar = "# ";
            break;
        }
    }
    // setup the argument parser
    ArgumentParser parser;
    ConfigArgs configArgs(parser);
    parser.setIgnoreUnknownArguments(false);
    // parse command line arguments
    try {
        parser.parseArgs(argc, argv);
    } catch (Failure &e) {
        cerr << shchar << "Unable to parse arguments: " << e.what() << endl;
        return 2;
    }
    try {
        // load configuration
        Config config;
        config.loadFromConfigFile(configArgs);
        config.loadFromArgs(configArgs);
        if(find_if(parser.mainArguments().cbegin(), parser.mainArguments().cend(), [&configArgs] (const Argument *arg) {
                   return arg != &configArgs.helpArg && arg->isPresent();
                }) != parser.mainArguments().cend()) {
            // create app
            QCoreApplication application(argc, argv);
            // setup manager
            Manager manager(config);
            cerr << shchar << "Loading databases ..." << endl;
            manager.registerDataBasesFromPacmanConfig();
            manager.registerDatabasesFromRepoIndexConfig();
            manager.initAlpmDataBases(configArgs.serverArg.isPresent());
            cerr << shchar << "Restoring cache ..." << endl;
            manager.restoreCache();

            if(configArgs.serverArg.isPresent()) {
                // setup the server
                Server server(manager, manager.config());
                manager.setAutoCacheMaintenanceEnabled(true);
                QObject::connect(&server, &Server::closed, &application, &QCoreApplication::quit);
                return application.exec();
            } else if(configArgs.buildOrderArg.isPresent()) {
                BuildOrderResolverCli resolver(manager, configArgs.buildOrderArg.values(), configArgs.addSourceOnlyDeps.isPresent());
                return resolver.exec();
            } else if(configArgs.mingwBundleArg.isPresent()) {
                MingwBundle bundle(manager, configArgs.mingwBundleArg.values(), configArgs.iconThemesArg.values(), configArgs.extraPackagesArg.values());
                bundle.createBundle(configArgs.targetDirArg.isPresent() ? configArgs.targetDirArg.values().front() : string("."),
                                    configArgs.targetNameArg.values().front(),
                                    configArgs.targetFormatArg.isPresent() ? configArgs.targetFormatArg.values().front() : string("zip"),
                                    configArgs.defaultIconThemeArg.isPresent() ? configArgs.defaultIconThemeArg.values().front() : string());
            } else if(configArgs.upgradeLookupArg.isPresent()) {
                UpgradeLookupCli upgradeLookup(manager, configArgs.upgradeLookupArg.values().front());
                return application.exec();
            }
        } else if(!configArgs.helpArg.isPresent()) {
            if(useShSyntax) {
                cerr << "export REPOINDEX_ERROR='No command line arguments specified. See --help for available commands.'" << endl;
            } else {
                cerr << "No command line arguments specified. See --help for available commands." << endl;
            }
        }
    } catch (const exception &ex) {
        Utilities::printError(ex);
        return 1;
    }
}
